/* -*- Mode: Javascript; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * Copyright (c) 2006, 2007 ActiveState Software Inc.
 * Refer to the license in the Komodo IDE installation
 * for the licensing information of this source code.
 *
 * winger-browser.js -- functions for using/accessing browser features
 */

if (typeof(JSDBGP) == 'undefined') {
    var JSDBGP = {};
}

JSDBGP.browser = {};

JSDBGP.browser.cache = {
    getResponseHeaders: function(cacheEntryDescriptor)
    {
        if (!cacheEntryDescriptor)
            return null;
        return cacheEntryDescriptor.getMetaDataElement("response-head");
    },
    
    getCacheEntry: function(/*nsIURI*/ uri, mode)
    {
        if (typeof(mode) == 'undefined')
            mode = Components.interfaces.nsICache.ACCESS_READ;
        
        var service = Components.classes["@mozilla.org/network/cache-service;1"].getService(Components.interfaces.nsICacheService);
        var session = service.createSession(uri.scheme.toUpperCase(), 0, true);
        session.doomEntriesIfExpired = false;

        return session.openCacheEntry(uri.spec, mode, false);
    },
    /**
     * getContent
     *
     * given a uri, look in mozilla's cache and see if we can get the content
     * for the uri, otherwise return null or raise exception on errors
     */
    getContent: function(uri) {
        try {
            var cacheEntry = JSDBGP.browser.cache.getCacheEntry(uri);
            if (!cacheEntry) {
                return null;
            }
            var cacheData = null;
            var headers = JSDBGP.browser.cache.getResponseHeaders(cacheEntry);
            var encoding = /^Content-Encoding:\s*(.*?)\s*(?:\;|$)/mi.exec(headers);
            if (encoding && encoding[1].length > 0) {
                cacheData = JSDBGP.browser. cache.uncompressCacheEntryData(cacheEntry, encoding[1]);
            }
            if (!cacheData && cacheEntry.isStreamBased()) {
                cacheData = JSDBGP.browser.cache.readCacheEntryData(cacheEntry, true);
            }
            return cacheData;
        } catch(ex) {
            dump("winger-browser.js: getContent: " + ex + "\n");
            return null;
        }
    },

    inCache: function(uri) {
        try {
            var cacheEntry = JSDBGP.browser.cache.getCacheEntry(uri);
            if (!cacheEntry) {
                return false;
            }
            cacheEntry.close();
            return true;
        } catch(ex) {
            return false;
        }
    },

    /**
     * uncompressCacheEntryData
     *
     * get all the data for a cached url and return it uncompressed
     */
    uncompressCacheEntryData: function(cacheEntry, contentEncoding) {
        var streamListener = {
            data : "",
            onDataAvailable : function(request, context, stream, offset, count) {
                this.data = this.data + JSDBGP.browser.cache.readStream(stream, true);
            },
            onStartRequest : function(request, context) {},
            onStopRequest : function(request, context, statusCode) {}
        }
        var converter = Components.classes["@mozilla.org/streamconv;1?from=" + contentEncoding + "&to=uncompressed"].
                                createInstance(Components.interfaces.nsIStreamConverter);

        try {
            var stream = cacheEntry.openInputStream(0);
     
            var uri = Components.classes["@mozilla.org/network/simple-uri;1"].
                            createInstance(Components.interfaces.nsIURI);
            uri.scheme = "http://gunzip";

            var channel = Components.classes["@mozilla.org/network/input-stream-channel;1"].
                            createInstance(Components.interfaces.nsIInputStreamChannel);
            channel.setURI(uri);
            channel.contentLength = -1;
            channel.contentType = contentEncoding;
            channel.contentStream = null;
     
            var request = channel.QueryInterface(Components.interfaces.nsIRequest);
            converter.asyncConvertData(contentEncoding, "uncompressed", streamListener, null);
            converter.onStartRequest(request, null);
            try {
                converter.onDataAvailable(request, null, stream, 0, stream.available());
            } finally {
                converter.onStopRequest(request, null, {});
            }
        } finally {
            cacheEntry.close();
        }
        return streamListener.data;
    },

    /**
     * readCacheEntryData
     *
     * get all the data for a cached url
     */
    readCacheEntryData: function(cacheEntry, async) {
        var inputStream = cacheEntry.openInputStream(0);
        var data = "";
        try {
            data = JSDBGP.browser.cache.readStream(inputStream, async);
        } finally {
            cacheEntry.close();
        }
        return data;
    },

    /**
     * readStream
     *
     * read all available data on the given stream
     */
    readStream: function(inputStream, async) {
        var stream = Components.classes["@mozilla.org/scriptableinputstream;1"].createInstance(Components.interfaces.nsIScriptableInputStream);
        stream.init(inputStream);

        var data = "";
        try {
            if (async) {
                var size = stream.available();
                var ch = null;
                for (var i = 0; i < size; i++) {
                   ch = stream.read(1);
                   ch ? data += ch : data += '\0';
                }
            } else {
                data = stream.read(stream.available());
            }
        } finally {
            stream.close();
        }
        return data;
    }
}

