/* -*- Mode: JavaScript; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * Copyright (c) 2006, 2007 ActiveState Software Inc.
 * Refer to the license in the Komodo IDE installation
 * for the licensing information of this source code.
 *
 * winger-properties.js -- manage properties
 */

if (typeof(JSDBGP) == 'undefined') {
    var JSDBGP = {};
}

JSDBGP.Properties = function(mainObj) {
    this.LocalVars = 0;
    this.GlobalVars = 1;
    this.InstanceVars = 2;
    this.contextProperties = {'locals' : this.LocalVars,
                              'globals' : this.GlobalVars,
                              'instance' : this.InstanceVars
    };
    this.contextPropertyNames = [];
    for (var k in this.contextProperties) {
        this.contextPropertyNames.push(k);
    }
    // It would be much better if we could tie these back to
    // their parent objects, but we don't always have that kind of info.
    //
    // Some, like canFindAgain, have a getter in browser.xul that
    // has an error in it when invoked in an unusual context, like this one.
    // 
    this.skippedNames = { 'constructor' : null,
                          'type' : null,
                          'domConfig' : null,
                          'canFindAgain' : null,
                          'selectionStart' : null,
                          'selectionEnd' : null,
                          // Shane reported these
                          'Component' : null,
                          'accessible' : null,
                          'popupBoxObject' : null,
                          'autoPosition' : null,
                          'boxObject' : null,
                          'controllers' : null,
                          'defaultView' : null,
                          'view' : null,
                          '' : null };
    this._skippedPropertyNames = { 'Packages' : null,
                                   'java' : null,
                                   'sun' : null,
                                   '' : null };
    this._skippedPropertyPatterns = [/^jslib[_CIR](?:\w|$)/ ];
    this.default_encoding = "base64";
    this.mainObj = mainObj;
    this.log = JSDBGP.Logging.getLogger("winger.properties");
    this.log.setLevel(JSDBGP.Logging.DEBUG);
};

// typeof(item) = {name:name, isWrapped:value, value: jsdValue or jsval}
JSDBGP.Properties.prototype.expandTopLevelProperty = function(item, hide_hidden, max_data_size) {
    if (!item.isWrapped) {
        return this.expandTopLevelValue(item, hide_hidden, max_data_size);
    }
    var property = <property />;
    var name = item.name;
    var jsdValue = item.value;






    var rawValue;
    property.@fullname = name;
    property.@name = name;
    property.@encoding = this.default_encoding;
    var xval = "";
    var numchildren = 0;
    try {

        var jsdType = jsdValue.jsType;
        switch(jsdType) {
        case jsdValue.TYPE_BOOLEAN:
        case jsdValue.TYPE_DOUBLE:
        case jsdValue.TYPE_INT:
        case jsdValue.TYPE_STRING:
            xval = jsdValue.getWrappedValue();
            property.@type = typeof(xval);
            xval = xval.toString()
            break;
        
        case jsdValue.TYPE_VOID:
            property.@type = "undefined";
            break;
        
        case jsdValue.TYPE_NULL:
            property.@type = "object";
            break;
        

        case jsdValue.TYPE_FUNCTION:
            property.@type = "function";
            break;

        case jsdValue.TYPE_OBJECT:
            if (name in this._skippedPropertyNames) {
                // Trying to do anything about these might hose the debugger.
                property.@type = "object";
            } else {
                var jsval = jsdValue.getWrappedValue();
                var retObj = {};
                this._finishObjectDetails(jsval, property, hide_hidden, max_data_size, retObj, false);
                if ('numchildren' in retObj) {
                    numchildren = retObj.numchildren;
                }
                if ('value' in retObj) {
                    xval = retObj.value;
                }
            }
            break;

        default:
            this.log.debug("expandTopLevelProperty: type(" + name
                           + ") = " + jsdType);
            property.@type = "unknown";
            break;
        }
    } catch(ex) {



        property.@type = "exception";
    }
    return this._finishProperty(property, xval, max_data_size, numchildren);
};

// typeof(item) = {name:name, isWrapped:value, value: jsdValue or jsval}
JSDBGP.Properties.prototype.expandTopLevelValue = function(item, hide_hidden, max_data_size) {
    var property = <property />;
    var name = item.name;
    var jsval = item.value;
    var stringify = ('stringify' in item ? item.stringify : false);
    var rawValue;
    property.@fullname = name;
    property.@name = name;
    property.@encoding = this.default_encoding;
    var xval = "";
    var numchildren = 0;
    try {
        var jsType = typeof(jsval);
        property.@type = jsType;
        switch (jsType) {
        case "number":
        case "boolean":
            xval = jsval.toString();
            break;
            
        case "string":
            xval = jsval;
            break;
            
        case "xml":
            xval = jsval.toXMLString();
            break;

        case "function":
            if (stringify) {
                xval = jsval.toString();
            }
            break;

        case "undefined":
            break;

        case "object":
            if (name in this._skippedPropertyNames) {
                // Trying to do anything about these might hose the debugger.
            } else {
                var retObj = {};
                this._finishObjectDetails(jsval, property, hide_hidden, max_data_size, retObj, stringify);
                if ('numchildren' in retObj) {
                    numchildren = retObj.numchildren;
                }
                if ('value' in retObj) {
                    xval = retObj.value;
                }
            }
            break;

        default:
            this.log.debug("expandTopLevelValue: type(" + name
                           + ") = unknown");
            property.@type = "unknown";
            break;
        }
    } catch(ex) {



        property.@type = "exception";
    }
    return this._finishProperty(property, xval, max_data_size, numchildren);
}

JSDBGP.Properties.prototype._finishObjectDetails = function(jsval, property, hide_hidden, max_data_size, retObj, stringify) {
    if (jsval) {
        var isArray = this.isArray(jsval);
        property.@type = isArray ? "array" : "object";
        var isWrapped = false;
        var countingOnly = true;
        retObj.numchildren = this.walkProperties(jsval, isWrapped, hide_hidden, countingOnly);
        try {
            property.@classname = jsval.constructor.name;
        } catch(ex2) {}
        // Go only two levels deep with arrays.  It's a recursion
        // guard.  No it's a timeout guard.  No it's two, two,
        // two guards in one.
        var adjusted_max_data_size = (retObj.numchildren <= 1
                                      ? max_data_size
                                      : max_data_size/retObj.numchildren);
        var valueValue = this._checkGetValueContents(jsval, adjusted_max_data_size, 5, hide_hidden, stringify);
        if (valueValue != null) {
            retObj.value = valueValue;
        }
    } else {
        property.@type = "object";
        retObj.value = "null";
    }
};

JSDBGP.Properties.prototype._finishProperty = function(property, xval, max_data_size, numchildren) {
    property.@size = xval.length;
    if (xval && xval.length > max_data_size) {
        if (max_data_size <= 3) {
            // Boundary condition possible with a big array
            xval = "...";
        } else {
            // Add ellipse to show we truncated
            xval = xval.substr(0, max_data_size - 4) + "...";
        }
    }
    if (numchildren > 0) {
        property.@children = "1";
        property.@numchildren = numchildren;
    } else {
        property.@children = "0";
    }
    if (typeof(xval) == "string" && xval.length > 0) {
        property.appendChild(this.mainObj.utils.base64Encode(xval));
    } else if (typeof(xval) == "xml") {
        var xstring = xval.toXMLString();
        if (xstring.length > 0) {
            property.appendChild(this.mainObj.utils.base64Encode(xstring));
        }
    }
    return property;
};

JSDBGP.Properties.prototype._extendPropertyName = function(parentName, curr_name) {



    if (/^[0-9][0-9]*$/.test(curr_name)) {
        return parentName + "[" + curr_name + "]";
    } else if (/[^a-zA-Z0-9_]/.test(curr_name)) {
        return parentName + "[\"" + curr_name.replace(/\\/g, "\\\\").replace(/\"/g, '\\\"') + "\"]";
    } else {
        return parentName + "." + curr_name;
    }
};

JSDBGP.Properties.prototype.getChildProperty = function(fullParentName, childName, raw_value, max_data_size, hide_hidden) {
    var property = this.expandTopLevelValue({name:childName, value:raw_value},
                                            hide_hidden, max_data_size);
    if (property) {
        property.@fullname = this._extendPropertyName(fullParentName, childName);
    }
    return property;
};

// returns a count or an array of [name, isWrapped, value] objects

JSDBGP.Properties.prototype.walkProperties = function(currVal, isWrapped, hide_hidden, counting, max_num_to_return) {
    if (arguments.length <= 3) {
        max_num_to_return = 0;



        if (arguments.length <= 2) {
            counting = false;



            if (arguments.length <= 1) {



                hide_hidden = true;
            }
        }
    }
    var resh = {}; // track the names
    var resa;
    if (!counting) {
        resa = [];
    }
    var jsval;
    if (counting) {
        if (max_num_to_return > 0) {



        }
        max_num_to_return = 0;
    }
    var numCollected = 0;
    if (isWrapped) {
        var listValue = {value: null};
        var lengthValue = {value: 0};
        currVal.getProperties(listValue, lengthValue);
        for (var i = 0; i < lengthValue.value; ++i) {
            var prop = listValue.value[i];
            var name = prop.name.stringValue;
            resh[":" + name] = 1;
            if (prop.flags == 0) {
                // These values are often dangerous if we try to eval
                // them.  Some of them are Rhino objects, the getters
                // seem to cross the js/java boundary and hose the debugger.
                continue;
            } else if (hide_hidden && this.mainObj.utils.is_hidable_name(name)) {
                continue;
            } else if (name in this.skippedNames) {
                continue;
            }
            var jsdValue = prop.value;
            if (hide_hidden && (jsdValue.jsType == jsdValue.TYPE_VOID
                                || jsdValue.jsType == jsdValue.TYPE_FUNCTION)) {
                continue;
            }
            numCollected += 1;
            if (!counting) {
                resa.push({name:name, isWrapped:true, value:jsdValue});
                if (max_num_to_return > 0 && numCollected >= max_num_to_return) {
                    break;
                }
            }
        }
        jsval = currVal.getWrappedValue();
    } else {
        jsval = currVal;
    }
    if (max_num_to_return == 0 || numCollected < max_num_to_return) {
        try {
            for (var jsProp in jsval) {
                if ((":" + jsProp) in resh) {
                    // No need to add new names to this hash.  We just want to
                    // skip any names found unsafe in the property list.
                    continue;
                }
                if (hide_hidden && this.mainObj.utils.is_hidable_name(jsProp)) {
                    continue;
                } else if (jsProp in this.skippedNames) {
                    continue;
                }
                     
                var tmp;
                try {
                    tmp = jsval[jsProp];
                } catch(ex) {







                    // Simulate the undefined value
                    tmp = this.undefined;
                }
                if (hide_hidden && (typeof(tmp) == "function"
                                    || typeof(tmp) == "undefined")) {
                                        // Too much info for DOM things



                                        continue;
                                    }
                numCollected += 1;
                if (!counting) {
                    resa.push({name:jsProp, isWrapped:false, value:tmp});
                    if (max_num_to_return > 0 && numCollected >= max_num_to_return) {
                        break;
                    }
                }
            }
        } catch(ex) {
            var show_exception;



            show_exception == ex.message.indexOf('Security error') == -1;

            if (show_exception) {
                this.log.exception(ex);
            }
        }
    }
    if (!counting) {
        // 3 criteria for array-ness
        var sorter;
        if (!isWrapped && this.isArray(currVal) && this._allPropsAreNumeric(currVal)) {
            sorter = this._cmpItemsByNumericIdx;
        } else {
            sorter = this._cmpItemsByStringIdx;
        }
        resa.sort(sorter);
        return resa;
    } else {
        return numCollected;
    }
};

// Misc helpers
JSDBGP.Properties.prototype.isArray = function(thing) {
    return (thing
            && typeof(thing) == 'object'
            && thing.constructor
            && thing.constructor.name == "Array");
};

JSDBGP.Properties.prototype._allPropsAreNumeric = function(currVal) {
    for (var prop in currVal) {
        if (isNaN(parseInt(prop))) {
            return false;
        }
    }
    return true;
};

// If it has a non-null value property, return its contents
// If it's an array of value-like things, 
// return either the bracketed list, recursively, if all items
// are displayable, or don't show anything.

// We're deliberately not showing all objects because evaluating
// all of them in a (for prop in obj) loop can trigger problems if
// some of the objects have getters that either throw errors or
// cross multi-tech boundaries, particularly into Rhino.
//
// This results in some anomalies:
// [{value:1}, {value:"2"}] ==> <<[1, "2"]>>
// but
// [{value:1}, {valuex:"2"}] ==> <<blank>>
//
// We assume value properties are always safe to evaluate.

JSDBGP.Properties.prototype._checkGetValueContents = function(jsval, data_size_left, levels, hide_hidden, stringify) {
    try {
        var tmp;
        switch (typeof(jsval)) {
        case "number":
        case "boolean":
            return jsval.toString();

        case "string":
            tmp = this._prepareString(jsval);
            break;

        case "xml":
            tmp = this._prepareString(jsval.toXMLString());
            break;

        case "function":
            return stringify ? jsval.toString() : null;
            
        case "undefined":
            return "undefined";

        case "object":
            if (!jsval) {
                return "null";
            } else if (this.isArray(jsval)) {
                if (levels <= 0) {
                    return "[...]";
                } else {
                    tmp = "[";
                    data_size_left -= 1;
                    var sawFirst = false;
                    var len = jsval.length;
                    var keep_adding = true;
                    for (var i = 0; i < jsval.length; i++) {
                        var innerVal = this._checkGetValueContents(jsval[i], data_size_left - 2, levels - 1, hide_hidden, stringify);
                        if (innerVal == null) {
                            return null;
                        }
                        if (!keep_adding) continue;
                        if (sawFirst) {
                            tmp += ', ';
                            data_size_left -= 2;
                        } else {
                            sawFirst = true;
                        }
                        tmp += innerVal;
                        if (data_size_left <= 1 && i < jsval.length) {
                            // If the previous item was a string val that
                            // couldn't fit, putting this in will result in
                            // [val1, val2, ..., ...]
                            
                            if (innerVal != "...") {
                                tmp += ", ...";
                            }
                            keep_adding = false;
                        } else {
                            data_size_left -= innerVal.length;
                        }
                    }
                    return tmp + "]";
                }
            } else if (stringify) {
                try {
                    var resa = [];
                    for (var jsProp in jsval) {
                        if (jsProp in this.skippedNames) {
                            continue;
                        } else if (hide_hidden && this.mainObj.utils.is_hidable_name(jsProp)) {
                            continue;
                        }
                        try {
                            var res = jsval[jsProp];
                            if (hide_hidden && (typeof(res) == "function"
                                                || typeof(res) == "undefined")) {
                                continue;
                            }
                            resa.push("'" + jsProp + "':" + res);
                        } catch(ex1) {



                        }
                    }
                    tmp = "{" + resa.join(", ") + "}";
                } catch(ex2) {



                    tmp = ex2.toString();
                }
                return tmp;
            } else if ('value' in jsval) {
                tmp = this._checkGetValueContents(jsval.value, data_size_left, 0, hide_hidden, stringify);
                if (typeof(tmp) != "string") {
                    return tmp;
                }
            } else {
                return null;
            }
            break;

        default:
            this.log.debug("unknown type: " + typeof(jsval));
            return null;
        }
    } catch(ex) {



        return null;
    }






    if (tmp.length > data_size_left) {
        tmp = tmp.substr(0, data_size_left) + "...";
        if (tmp.substr(0, 1) == '"') {
            tmp += '"';
        }
    }
    return tmp;
};

JSDBGP.Properties.prototype._cmpItemsByNumericIdx = function(a, b) {
    var a1 = parseInt(a.name);
    var b1 = parseInt(b.name);
    return a1 < b1 ? -1 : (a1 == b1 ? 0 : 1);
}

JSDBGP.Properties.prototype._cmpItemsByStringIdx = function(a, b) {
    return a.name < b.name ? -1 : (a.name == b.name ? 0 : 1);
}

JSDBGP.Properties.prototype._prepareString = function(str) {
    return ('"'
            + str.replace(/\\/g, "\\\\").replace(/\"/g, '\\\"')
            + '"');
};

JSDBGP.Properties.prototype._matchesSkipablePropertyPatterns = function(name) {
    for (var i in this._skippedPropertyPatterns) {
        var ptn = this._skippedPropertyPatterns[i];
        if (ptn.test(name)) {
            return true;
        }
    }
    return false;
}
