/* -*- Mode: JavaScript; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * Copyright (c) 2006, 2007 ActiveState Software Inc.
 * Refer to the license in the Komodo IDE installation
 * for the licensing information of this source code.
 *
 * winger-utils.js -- static utility functions
 */

try {
include('chrome://jslib/content/io/filesystem.js');
include('chrome://jslib/content/io/chromeFile.js');
include('chrome://jslib/content/io/file.js');
include('chrome://jslib/content/io/fileUtils.js');

if (typeof(JSDBGP) == 'undefined') {
    var JSDBGP = {};
}

JSDBGP.Utils = function(mainObj) {
    this.mainObj = mainObj;
    this.canonicalize_uris = {};
    this.log = JSDBGP.Logging.getLogger("winger.utils");
    this.log.setLevel(JSDBGP.Logging.DEBUG);

    this._unc_path_re = /^\\\\(.+?)\\(.*)$/;
    this._win_path_re = /^(\w:[/\\].*)$/;
    this._abs_path = /^([\\/].*)$/;
    this._all_caps_prop_name_re = /^[_A-Z][A-Z0-9_]+$/;

    this.nsIOService = (Components.classes["@mozilla.org/network/io-service;1"]
                      .getService(Components.interfaces.nsIIOService));

    this.js_dbgp_javascript_url_re = 
    new RegExp('^dbgp://js/\\d+/<javascript-url>:(\\d+)\\.js$');
    this.js_dbgp_with_base_re =
    new RegExp('^dbgp://js/(.*)/\\d+/.*:(\\d+)\\.js$');
    this.js_dbgp_no_base_re =
    new RegExp('^dbgp://js/\\d+/(.*?)(?:<index>)?:(\\d+)\\.js$');
    this.dbgp_http_name_re = 
    new RegExp('^dbgp:///[a-z]');

    this.validURLS = {};
    this._filtered_URLs = {};
    





    
    try {
        var fileProperties = (Components.classes["@mozilla.org/file/directory_service;1"]
                              .getService(Components.interfaces.nsIProperties));
        var profDirFile  = fileProperties.get("ProfD", Components.interfaces.nsIFile);
        var installationDirFile = fileProperties.get("CurProcD", Components.interfaces.nsIFile);
        var fileHandler = (Components.classes["@mozilla.org/network/io-service;1"]
                           .getService(Components.interfaces.nsIIOService)
                           .getProtocolHandler("file")
                           .QueryInterface(Components.interfaces.nsIFileProtocolHandler));
        this._profile_dir = fileHandler.newFileURI(profDirFile).spec;
        this._installation_dir = fileHandler.newFileURI(installationDirFile).spec;
    } catch(ex) {
        this.log.exception(ex, "Can't get profile dir");
        this._profile_dir = null;
        this._installation_dir = null;
    }
    
    this.getURI = function(rawUri) {
        try {
            const nsIStandardURL = Components.interfaces.nsIStandardURL;
            var url = Components.classes["@mozilla.org/network/standard-url;1"].createInstance(nsIStandardURL);
            url.init(nsIStandardURL.URLTYPE_STANDARD, null, rawUri, null, null);
            return url.QueryInterface(Components.interfaces.nsIURI);
        } catch(ex) {
            this.log.exception(ex, "utils.getURI(" + rawUri + "): ");
            return {spec:null};
        }
    };

    // JavaScript internally uses file:/path URIs, instead of file://<host>/...
    // and mucks up windows drive letters too.
    
    this.canonicalize_uri = function(uri) {
        if (!uri) {
            return "";
        }
        var canonical_uri;
        if (uri in this.canonicalize_uris) {



            return this.canonicalize_uris[uri];
        } else {
            // Assign and return
            return (this.canonicalize_uris[uri] = this._canonicalize_uri_aux(uri));
        }
    };

    this._canonicalize_uri_aux = function(uri) {
        if (uri.indexOf("javascript:") == 0) {
            return uri;
        }







        var uri1 = this.getURI(uri);
        if (!uri1.spec) {
            // It's a filename of some kind
            // Cases:
            // UNC paths: \\<host>\path
            // Windows file: <letter>:[/\]path
            // Unix file: /path

            // Don't bother trying to determine which platform we're on.
            // Just do the pattern-matching
        
            var matchObj;
            if (this._unc_path_re.test(uri)) {
                throw new Error("UNC paths like " + uri + " aren't supported");
            } else if ((matchObj = this._win_path_re.exec(uri))) {
                uri1 = this.getURI("file:///" + matchObj[0]);



            } else if (this._abs_path.test(uri)) {
                uri1 = this.getURI("file:///" + uri);



            }
        }
        if (!uri1 || !uri1.spec) {
            throw new Error("Can't canonicalize filename {" + uri + "}");
        } else {







            return uri1.spec;
        }
    };

    // Taken from venkman-utils.js

    this.dumpObjectTree = function(o, recurse, compress, level) {
        var s = "";
        var pfx = "";

        if (typeof recurse == "undefined") {
            recurse = 0;
        }
        if (typeof level == "undefined") {
            level = 0;
        }
        if (typeof compress == "undefined") {
            compress = true;
        }

        var i;
        for (i = 0; i < level; i++) {
            pfx += (compress) ? "| " : "|  ";
        }

        var tee = (compress) ? "+ " : "+- ";

        if (typeof o == "string") {
            s += pfx + tee + " (string) ";
            if (o.length > 200) {
                s += o.substr(0, 40) + "... " + o.length + " chars\n";
            } else {
                s += o + "'\n";
            }
        } else {
            for (i in o) {
                var t;
                try {
                    t = typeof o[i];
        
                    switch (t) {
                        case "function":
                            var sfunc = String(o[i]).split("\n");
                            if (sfunc[2] == "    [native code]")
                                sfunc = "[native code]";
                            else
                                sfunc = sfunc.length + " lines";
                            s += pfx + tee + i + " (function) " + sfunc + "\n";
                            break;
                    
                        case "object":
                            s += pfx + tee + i + " (object) " + o[i] + "\n";
                            if (!compress)
                                s += pfx + "|\n";
                            if ((i != "parent") && (recurse))
                                s += this.dumpObjectTree(o[i], recurse - 1,
                                                         compress, level + 1);
                            if (o[i] == "[xpconnect wrapped jsdIValue]" && 'getWrappedValue' in o[i]) {
                                s += this.dumpObjectTree(o[i].getWrappedValue(), recurse - 1,
                                                         compress, level + 1);
                            }
                            break;
                    
                        case "string":
                            if (o[i].length > 200)
                                s += pfx + tee + i + " (" + t + ") " + 
                                    o[i].length + " chars\n";
                            else
                                s += pfx + tee + i + " (" + t + ") '" + o[i] + "'\n";
                            break;
                    
                        default:
                            s += pfx + tee + i + " (" + t + ") " + o[i] + "\n";
                        }
                } catch (ex) {
                    s += pfx + tee + i + " (exception) " + ex + "\n";
                }

                if (!compress) {
                    s += pfx + "|\n";
                }
            }
        }

        s += pfx + "*\n";
    
        return s;
    
    };

    // More stuff from venkman
    this.formatFrame = function(f) {
        var str;
        try {
            var url = (f.isNative) ? "<native>" : f.script.fileName;
            str = url;
            var line = f.line;
            var pc = f.pc;
            str += "/tag " + f.script.tag
                str += "/line " + line;
            str += "/pc " + pc;
            str += ("(PCMAP_SOURCETEXT(pc) = "
                    + f.script.pcToLine(pc, PCMAP_SOURCETEXT)
                    + ", PCMAP_PRETTYPRINT(pc) = "
                    + f.script.pcToLine(pc, PCMAP_PRETTYPRINT));
            return str;
        } catch(ex) {
            this.log.exception(ex);
            return "Can't get frame";
        }
    };

    // Filters

    this.isURLDBGP = function(url) {
        return (url.indexOf("chrome://jsdbgp/content/winger-") == 0
                || url.indexOf("chrome://jsdbgp/components/jsdbgp-") == 0);
    };

    // Need to add these.

    this.isURLFiltered = function(url) {



        var res = this._isURLFiltered_aux(url);


















        return res;
    };

    this._isURLFiltered_aux = function(url) {
        if (!url) {



            // Filter out null URLs
            return true;
        } else if (url.indexOf(":") == -1) {
            // Handle URIs like XStringBundle




            return true;
        }
        try {
            var nsURI = this.nsIOService.newURI(url, null, null);
            var spec = nsURI.spec;
        } catch(ex) {




            return true;
        }
        // Check for JavaScript URIs
        if (nsURI.scheme == "javascript") {
            // We always want to step into these.
            return false;
        } else if (!nsURI.scheme) {
            return true;
        } else if (!this.mainObj.console.isXulApp) {
            if (nsURI.scheme == "chrome") {
                if (spec.indexOf(this.mainObj.console.parent_dir) >= 0) {
                    return false;
                } else {
                    return true;
                }
            } else if (nsURI.scheme == "file") {
                if (this._profile_dir && spec.indexOf(this._profile_dir) >= 0) {
                    // Some extensions live in profile space.
                    // Note that this._profile_dir has had the leading
                    // 'file:///' chopped off because JSD reports file URIs
                    // with only one slash.
                    return true;
                } else if (this._installation_dir && spec.indexOf(this._installation_dir) >= 0) {
                    // Sometimes we end up stopping in non-chrome files that
                    // live in the installation directory.
                    // 'file:///' chopped off because JSD reports file URIs
                    // with only one slash.
                    return true;
                } else {
                    return false;
                }
            }
        }
            
        // Accept everything else
        return false;
    };

    
    this.cantGetSource = function(url) {
        if (url.indexOf("?") == -1) {



            return false;
        } else if (url in this.validURLS) {



            return false;
        } else if (JSDBGP.browser.cache.inCache(this.getURI(url))) {
            // Keep trying until we hit



            this.validURLS[url] = null;
            return false;
        } else {



            return true;
        }
    }

    this.Mapped_URI_Matches = function(currFileName, targetFileName) {







        if (this.canonicalize_uri(currFileName) == this.canonicalize_uri(targetFileName)) {
            return true;
        }
        return false;
    };

    this.pathToURI = function(local_path) {
        try {
            var utils = new FileUtils();
            var uri = utils.pathToURL(local_path);
            return this.canonicalize_uri(uri);
        } catch(ex) {



        }
        return local_path;
    };

    // [...the base64] code was written by Tyler Akins and has been placed in the
    // public domain.  It would be nice if you left this header intact.
    // Base64 code from Tyler Akins -- http://rumkin.com

    this.keyStr = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";

    this.base64Encode = function (input) {
        var output = "";
        var chr1, chr2, chr3;
        var enc1, enc2, enc3, enc4;
        var i = 0;

        do {
            chr1 = input.charCodeAt(i++);
            chr2 = input.charCodeAt(i++);
            chr3 = input.charCodeAt(i++);

            enc1 = chr1 >> 2;
            enc2 = ((chr1 & 3) << 4) | (chr2 >> 4);
            enc3 = ((chr2 & 15) << 2) | (chr3 >> 6);
            enc4 = chr3 & 63;

            if (isNaN(chr2)) {
                enc3 = enc4 = 64;
            } else if (isNaN(chr3)) {
                enc4 = 64;
            }

            var keyStr = this.keyStr;
            output = output + keyStr.charAt(enc1) + keyStr.charAt(enc2) + 
                keyStr.charAt(enc3) + keyStr.charAt(enc4);
        } while (i < input.length);
   
        return output;
    };

    this.base64Decode = function(input) {
        var output = "";
        var chr1, chr2, chr3;
        var enc1, enc2, enc3, enc4;
        var i = 0;

        // remove all characters that are not A-Z, a-z, 0-9, +, /, or =
        input = input.replace(/[^A-Za-z0-9\+\/\=]/g, "");

        do {
            var keyStr = this.keyStr;
            enc1 = keyStr.indexOf(input.charAt(i++));
            enc2 = keyStr.indexOf(input.charAt(i++));
            enc3 = keyStr.indexOf(input.charAt(i++));
            enc4 = keyStr.indexOf(input.charAt(i++));

            chr1 = (enc1 << 2) | (enc2 >> 4);
            chr2 = ((enc2 & 15) << 4) | (enc3 >> 2);
            chr3 = ((enc3 & 3) << 6) | enc4;

            output = output + String.fromCharCode(chr1);

            if (enc3 != 64) {
                output = output + String.fromCharCode(chr2);
            }
            if (enc4 != 64) {
                output = output + String.fromCharCode(chr3);
            }
        } while (i < input.length);

        return output;
    };

    this.classifyURL = function(url) {
        if (this.js_dbgp_javascript_url_re.test(url)) {
            return JSDBGP.Constants.JavaScriptURL;
        } else if (this.js_dbgp_no_base_re.test(url)) {
            return JSDBGP.Constants.JavaScriptIndex;
        } else if (this.js_dbgp_with_base_re.test(url)) {
            return JSDBGP.Constants.JavaScriptWithBase;
        } else if (this.dbgp_http_name_re.test(url)) {
            return JSDBGP.Constants.DBGP_HTTP_URL;
        } else {
            return JSDBGP.Constants.StandardURL;
        }
    };

    this.cleanString = function(str) {
        return str.replace(/\\n/g, "\n").
            replace(/\\r/g, "\r").
            replace(/\\t/g, "\t").
            replace(/\\(.)/g, "$1");
    };
    this.getFileSource = function(uri, resultObj) {
        var content = null;
        var status;
        // this.log.debug("utils.getFileSource(" + uri + ")");
        try {
            var reader = null;
            if (uri.indexOf("file:/") == 0) {
                var utils = new FileUtils();
                var localName = utils.urlToPath(uri);
                if (typeof(localName) == "string") {
                    reader = new File(localName);
                } else {
                    reader = new File(uri);
                }
            } else if (uri.indexOf("chrome:/") == 0) {
                reader = new ChromeFile(uri);
            }
        } catch(ex) {
            this.log.exception(ex, "utils getFileSource(" + uri + "): ");
        }
        if (reader) {
            if (reader.open(JS_FILE_READ_MODE) != JS_LIB_OK) {
                this.log.debug("Jslib open (" + uri + ") failed: ");
                reader.close();
                status = false;
            } else {
                content = reader.read();
                reader.close();
                status = true;
                resultObj.finalSource = content;
                // this.log.debug("getFileSource(" + uri + ") => {{" + content + "}}");
            }
        } else {
            // try the browser cache now
            try {
                content = JSDBGP.browser.cache.getContent(this.getURI(uri));
                if (!content) {
                    status = false;



                } else {
                    status = true;
                    resultObj.finalSource = content;
                }
            } catch(e) {
                // the uri is not in cache either!
                status = false;
                this.log.exception(e);
            }
        }
        if (!content) {



            status = false;
        }
        resultObj.status = status;
        return status;
    };
    
    this.is_hidable_name = function(propName) {
        if (this._all_caps_prop_name_re.test(propName)) {
            return true;
        }
        if (propName.charCodeAt(0) == "_".charCodeAt(0)) {
            return true;
        }
        return false;
    };
    









    // From MochiKit/Base.js (MIT Licensed)
    this.keys = function (obj) {
        /***

            Return an array of the property names of an object
            (in no particular order).

        ***/
        var rval = [];
        for (var prop in obj) {
            rval.push(prop);
        }
        return rval;
    };

    // Derived from the above func

    this.keysLength = function (obj) {
        var rval = 0;
        for (var prop in obj) {
            rval += 1;
        }
        return rval;
    };
    
    this.memberOf = function(item, list) {
        for (var i in list) {
            if (item == list[i]) {
                return true;
            }
        }
        return false;
    };

};
} catch(ex) {
    dump("Failed to load winger-utils.js: " + ex + "\n");
}
